#include <windows.h>
#include <shlwapi.h>
#include <shlobj.h>
#include <strsafe.h>
#include <shellapi.h>
#include <string.h>

#ifdef _DEBUG

#define CONCAT_IMP(x, y) x##y
#define CONCAT(x, y) CONCAT_IMP(x, y)

#define DBG_TEXT(string, ...) \
	do \
	{ \
		wprintf(L"DHeapInst.com: " CONCAT(L, string), __VA_ARGS__); \
		putchar('\n'); \
	} \
	while(0)

#else

#define DBG_TEXT(string, ...) ((void)0)

#endif

// returns the file or full path name of the module specified
// by the hMod parameter
LPWSTR GetModuleName(HMODULE hMod, bool bFileNameOnly)
{
	DWORD dwBufferSize = MAX_PATH;
	LPWSTR pwszFileName = NULL;
	LPWSTR pwszFilePath = new WCHAR[dwBufferSize];
	DWORD dwWrittenSize = 0;
	for(;;)
	{
		dwWrittenSize = GetModuleFileName(hMod, pwszFilePath, dwBufferSize);
		if(dwWrittenSize == dwBufferSize)
		{
			LPWSTR pTemp = new WCHAR[dwBufferSize *= 2];
			delete [] pwszFilePath;
			pwszFilePath = pTemp;
		}
		else break;
	}
	if(bFileNameOnly)
	{
		LPWSTR pwszFileNamePart = PathFindFileName(pwszFilePath);
		DWORD dwCharsRequired = static_cast<DWORD>((dwWrittenSize - (pwszFileNamePart - pwszFilePath)) + 1);
		pwszFileName = new WCHAR[dwCharsRequired];
		if(pwszFileName)
		{
			StringCchCopy(pwszFileName, dwCharsRequired, pwszFileNamePart);
		}
		delete [] pwszFilePath;
	}
	else
	{
		pwszFileName = pwszFilePath;
	}
	return pwszFileName;
}

LPWSTR GetExeVersionOfFile()
{
	LPWSTR pwszExeFilePath = GetModuleName(NULL, false);
	PathRenameExtension(pwszExeFilePath, L".exe");
	return pwszExeFilePath;
}

LPWSTR QuoteString(LPCWSTR pwszString)
{
	DWORD dwCharsInString = static_cast<DWORD>(wcslen(pwszString));
	DWORD dwCharLen = dwCharsInString + 3;
	LPWSTR pwszQuotedString = new WCHAR[dwCharLen];
	pwszQuotedString[0] = L'\"';
	StringCchCopy(pwszQuotedString + 1, dwCharLen - 1, pwszString);
	pwszQuotedString[dwCharLen - 2] = L'\"';
	pwszQuotedString[dwCharLen - 1] = 0;
	DBG_TEXT("Original string %s, quoted %s", pwszString, pwszQuotedString);
	return pwszQuotedString;
}

int FindSingleLetterArgument(WCHAR** argv, WCHAR letter)
{
	WCHAR wcUpperLetter = towupper(letter);
	letter = towlower(letter);
	for(int index = 0; argv[index]; ++index)
	{
		if(
			(argv[index][0] == L'-' || argv[index][0] == L'/') && 
			(argv[index][1] == letter || argv[index][1] == wcUpperLetter)
		)
		{
			DBG_TEXT("Found %c%c argument at position %d", argv[index][0], letter, index);
			return index + 1;
		}
	}
	return -1;
}

void LoadAndModifyRegistryParams(WCHAR** argv)
{
	LONG lRet;
	HKEY hKey = NULL;
	lRet = RegOpenKeyEx(
		HKEY_LOCAL_MACHINE,
		L"SYSTEM\\CurrentControlSet\\services\\DHEAPDMP\\parameters",
		REG_OPTION_NON_VOLATILE,
		KEY_QUERY_VALUE | KEY_SET_VALUE,
		&hKey);
	if(lRet == ERROR_SUCCESS)
	{
		LONGLONG llValues[2] = {0};
		DWORD dwBufferSize = sizeof(llValues);
		VALENT vaValues[2] = {{0}, {0}};
		vaValues[0].ve_valuename = L"BaseAddress";
		vaValues[1].ve_valuename = L"EResource";
		lRet = RegQueryMultipleValues(
			hKey,
			vaValues,
			ARRAYSIZE(vaValues),
			reinterpret_cast<LPWSTR>(llValues),
			&dwBufferSize);
		if(lRet == ERROR_SUCCESS)
		{
			LONGLONG llEResourceOffset = llValues[1] - llValues[0];
			lRet = RegSetValueEx(
				hKey,
				L"EResource",
				0,
				REG_QWORD,
				reinterpret_cast<BYTE*>(&llEResourceOffset),
				sizeof(LONGLONG)
				);
		}
		DBG_TEXT("RegSet/QueryValueEx returned %d", lRet);
		RegCloseKey(hKey);
	}
	else
	{
		DBG_TEXT("RegOpenKeyEx failed with error %d", lRet);
	}
}

DWORD LaunchRealInstaller()
{
	DWORD dwRealInstallerRet = STILL_ACTIVE;
	// first get the quoted path of the real installer
	LPCWSTR pwszRealInstallerPath = GetExeVersionOfFile();

	// find the args we were given
	LPCWSTR pwszCommandLine = GetCommandLine();
	LPWSTR pwszArgs = PathGetArgs(pwszCommandLine);

	// +1 for the seperating space, +1 for the NULL
	DWORD dwCommandAndArgsLen = wcslen(pwszRealInstallerPath) + wcslen(pwszArgs) + 2;
	LPWSTR pwszFullCmdLine = new WCHAR[dwCommandAndArgsLen];
	StringCchPrintf(pwszFullCmdLine, dwCommandAndArgsLen, L"%s %s", pwszRealInstallerPath, pwszArgs);

	LPWSTR pwszQuotedInstallerPath = NULL;
	HRESULT hr = SHEvaluateSystemCommandTemplate(pwszFullCmdLine, &pwszQuotedInstallerPath, &pwszArgs, NULL);
	delete [] pwszFullCmdLine;
	if(FAILED(hr))
	{
		DBG_TEXT("SHEvaluateSystemCommandTemplate returned 0x%x", hr);
		dwCommandAndArgsLen += 2;
		pwszFullCmdLine = new WCHAR[dwCommandAndArgsLen];
		StringCchPrintf(pwszFullCmdLine, dwCommandAndArgsLen, L"\"%s\" %s", pwszRealInstallerPath, pwszArgs);
		pwszArgs = pwszFullCmdLine;
		DBG_TEXT("Launching %s", pwszArgs);
	}
	else
	{
		DBG_TEXT("Launching %s with args \"%s\"", pwszQuotedInstallerPath, pwszArgs);
	}

	delete [] pwszRealInstallerPath;

	// start the real installer with the same args
	// and wait for it to finish, taking note of its return code
	STARTUPINFO si = {sizeof(si), 0};
	PROCESS_INFORMATION pi = {0};
	
	if(CreateProcess(pwszQuotedInstallerPath, pwszArgs, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
	{
		CloseHandle(pi.hThread);
		WaitForSingleObject(pi.hProcess, INFINITE);
		GetExitCodeProcess(pi.hProcess, &dwRealInstallerRet);
		DBG_TEXT("Real installer returned %d", dwRealInstallerRet);
		CloseHandle(pi.hProcess);
	}
	else
	{
		DBG_TEXT("CreateProcess failed with error %d", GetLastError());
	}
	// if pwszRealInstallerPath is valid then SHEvaluateSystemCommandTemplate succeeded
	// so use CoTaskMemFree to free it
	if(pwszQuotedInstallerPath)
	{
		CoTaskMemFree(pwszQuotedInstallerPath);
		CoTaskMemFree(pwszArgs);
	}
	// otherwise we allocated the memory
	else
	{
		delete [] pwszArgs;
	}
	return dwRealInstallerRet;
}

bool IsAnInstallationOperation(WCHAR** argv)
{
	return FindSingleLetterArgument(argv, 'r') == -1 && FindSingleLetterArgument(argv, 'h') == -1;
}

int __cdecl wmain(int argc, WCHAR** argv)
{	
	// te real installer returns 1 on failure and 0 on success
	DWORD dwRealInstallerRet = LaunchRealInstaller();
	if((dwRealInstallerRet == 0) && (IsAnInstallationOperation(argv)))
	{
		LoadAndModifyRegistryParams(argv);
	}
	return static_cast<int>(dwRealInstallerRet);
}
